package lcsb.mapviewer.web.api.project;

import lcsb.mapviewer.annotation.data.MeSH;
import lcsb.mapviewer.annotation.services.TaxonomySearchException;
import lcsb.mapviewer.annotation.services.annotators.AnnotatorException;
import lcsb.mapviewer.api.BaseController;
import lcsb.mapviewer.api.minervanet.MinervaNetController;
import lcsb.mapviewer.common.Pair;
import lcsb.mapviewer.model.MinervaEntity;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.persist.dao.map.ProjectProperty;
import lcsb.mapviewer.services.ObjectNotFoundException;
import lcsb.mapviewer.services.interfaces.IMeshService;
import lcsb.mapviewer.services.interfaces.IProjectService;
import lcsb.mapviewer.services.interfaces.ITaxonomyService;
import lcsb.mapviewer.services.interfaces.IUserService;
import lcsb.mapviewer.web.api.NewApiResponseSerializer;
import lcsb.mapviewer.web.api.ResponseDecorator;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.security.core.Authentication;
import org.springframework.validation.annotation.Validated;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

@RestController
@Validated
@RequestMapping(
    value = {
        "/minerva/new_api/projects"
    },
    produces = MediaType.APPLICATION_JSON_VALUE)
public class NewProjectController extends BaseController {

  private final Logger logger = LogManager.getLogger();

  private final MinervaNetController minervaNetController;
  private final IMeshService meshService;
  private final IUserService userService;
  private final IProjectService projectService;
  private final ITaxonomyService taxonomyService;
  private final NewApiResponseSerializer serializer;

  @Autowired
  public NewProjectController(
      final IMeshService meshService,
      final MinervaNetController minervaNetController,
      final IProjectService projectService,
      final IUserService userService,
      final ITaxonomyService taxonomyService,
      final NewApiResponseSerializer serializer) {
    this.meshService = meshService;
    this.userService = userService;
    this.projectService = projectService;
    this.minervaNetController = minervaNetController;
    this.serializer = serializer;
    this.taxonomyService = taxonomyService;
  }


  private final ResponseDecorator projectResponseDecorator = new ResponseDecorator() {

    @Override
    public List<Pair<String, Object>> getDecoratedValuesForRaw(final Object object) {
      return new ArrayList<>();
    }

    @Override
    public List<Pair<String, Object>> getDecoratedValues(final MinervaEntity object) {
      final List<Pair<String, Object>> result = new ArrayList<>();
      if (object instanceof Project) {
        final Project project = (Project) object;
        String diseaseName = null;
        if (project.getDisease() != null) {
          final MeSH mesh;
          try {
            mesh = meshService.getMesh(project.getDisease());
            if (mesh != null) {
              diseaseName = mesh.getName();
            }
          } catch (final AnnotatorException e) {
            logger.error("Problem with mesh", e);
          }
        }
        String organismName = null;
        if (project.getOrganism() != null) {
          try {
            organismName = taxonomyService.getNameForTaxonomy(project.getOrganism());
          } catch (final TaxonomySearchException e) {
            logger.error("Problem with taxonomy", e);
          }
        }

        boolean sharedInMinervaNet = false;
        try {
          sharedInMinervaNet = minervaNetController.getSharedProjects().contains(project.getProjectId());
        } catch (final Exception e) {
          logger.error("Problem with minerva net", e);
        }

        result.add(new Pair<String, Object>("sharedInMinervaNet", sharedInMinervaNet));
        result.add(new Pair<String, Object>("diseaseName", diseaseName));
        result.add(new Pair<String, Object>("organismName", organismName));
      }
      return result;
    }
  };

  @PreAuthorize("hasAnyAuthority('IS_ADMIN', 'READ_PROJECT:' + #projectId)")
  @GetMapping(value = "/{projectId:.+}")
  public ResponseEntity<?> getProject(final @PathVariable(value = "projectId") String projectId)
      throws ObjectNotFoundException {
    final Project project = projectService.getProjectByProjectId(projectId, true);
    if (project == null) {
      throw new ObjectNotFoundException("Project with given id doesn't exist");
    }
    return serializer.prepareResponse(project, projectResponseDecorator);

  }

  @GetMapping(value = "/")
  public ResponseEntity<?> getProjects(
      final Pageable pageable,
      final Authentication authentication
  ) {
    final User user = userService.getUserByLogin(authentication.getName(), true);
    final Map<ProjectProperty, List<?>> properties = new HashMap<>();

    properties.put(ProjectProperty.USER, Collections.singletonList(user));

    final Page<Project> page = projectService.getProjectsByFilter(properties, pageable, true);
    return serializer.prepareResponse(page, projectResponseDecorator);
  }

}